/******************************************************************************
metal HLSL
by Ben Cloward
bsmji@hotmail.com

This shader is intended to be used for metal surfaces.  It allows you to create
shiny metal, dull metal, and everthing in between and control those properties
per-pixel.

It requires hardware support for DirectX 9 and pixel shader 2.0.

It accepts the following inputs:

Ambient Color
Reflection Cube
Reflection Color
Diffuse Cube
Diffuse Color
Diffuse Texture
Shinniness Mask (alpha channel of diffuse texture)
Shininess
Normal Map Texture
Offset Amount
Point Light Position
Point Light Color

Ambient Color - 
A color value that controls the overall brightness of the whole object.  The RGB
values can be greater than one if you need them to be.

Reflection Cube -
This is a cube map that creates the reflection in the shiny areas of the surface.
Use a 16 bit floating point DDS cube map for best results.

Reflection Color -
This color is used to control the appearance of the reflection.  You can make it
brighter or duller and also add tint to it with this color.

Diffuse Cube -
This is a cube map that is diffusely convolved.  It's used to light the dull areas
of the surface.  See this tutorial for more information:
http://www.debevec.org/HDRShop/tutorial/tutorial6.html

Diffuse Color - 
This color is used to control the appearance of the diffuse lighing.  You can make
it brighter or duller and also add tint to it with this color.

Diffuse Texture - 
Use this 2d texture map to control the surface color appearance of your object.

Shinniness Mask (alpha channel of diffuse texture) - 
The alpha channel of the diffuse texture determines how shiny/reflective or dull
the surface is.  If the pixel in the alpha channel is white, that spot on your
object will act like a perfect mirror.  If the pixel is black, that spot on the
object will be very dull and diffuse.  Grey pixels in between will be a
combination of dull and shiny depending on how bright they are.  This allows you
to create metal that has veriable surface properties - like chrome that has
rusty spots.  You can make a brushed look by adding long thin streaks in the
alpha channel.

Shininess - 
This value controls the size of the specular highlight created by the point light.
A higher value creates a smaller, sharper highlight (more shiny) and a lower
value creates a larger, blurrier highlight.

Normal Map Texture - 
This shader supports normal mapping so you can add more detail to the surfaces of
your metal objects.  See my normal mapping tutorial for more details:
http://www.monitorstudios.com/bcloward/resources_tutorials.html

Offset Amount - 
The idea behind Offset Mapping is to "punch out" the surface of the model based on
a height map that goes along with the normal map so you get parallax movement on
the surface when your viewpoint changes.  The effect is achieved by multiplying the
UV coordinates used for the normal map and the diffuse map by a height map value
(stored in the alpha channel of the normal map) and by the view vector in the pixel
shader. We then use the resulting coordinates to look up our normal map and diffuse
map - instead of the original texture coordinates. You can control the height of
the bumps with the "Offset" value in the shader.  You also need to include a height
map in the alpha channel of your normal map.  If you created the normal map with
high res geometry, you can get a height map by rendering a Z depth image of the
geometry. If you created your normal map by converting a bump map, you can just use
the original bump map in the alpha channel. If you just have a normal map and need
to generate a height map from it, you can find a great tool to do that right here:
http://66.70.170.53/Ryan/heightmap/heightmap.html

Point Light Position and Color - 
This shader has support for one dynamic point light in addition to the global
illumination lighting.  You can choose which point light to use and also set the 
color of the light.
	
******************************************************************************/

/************* TWEAKABLES **************/

half4 ambient : Ambient
<
    string UIName = "Ambient Color";
> = {0.5f, 0.5f, 0.5f, 1.0f};

texture reflectCube : ReflectMap
<
    string name = "default_reflection.dds";
	string UIName = "Reflection Cube";
    string type = "Cube";
>;

half4 specularColor : Specular
<
    string UIName = "Reflection Color";
> = { 0.2f, 0.2f, 0.2f, 1.0f };

texture diffuseCube : EnvMap
<
    string name = "default_reflection.dds";
	string UIName = "Diffuse Cube";
    string type = "Cube";
>;

half4 surfColor : Diffuse
<
    string UIName = "Diffuse Color";
> = {0.5f, 0.5f, 0.5f, 1.0f};

texture colorTexture : DiffuseMap
<
    string name = "default_color.dds";
	string UIName = "Diffuse Texture";
    string TextureType = "2D";	
>;



half shininess <
	string UIWidget = "slider";
    float UIMin = 1;
    float UIMax = 128;
    string UIName = "Shininess";
> = 40;

texture normalMap : NormalMap
<
    string name = "default_bump_normal.dds";
	string UIName = "Normal Map";
    string TextureType = "2D";	
>;

half offset <
    float UIMin = -1;
    float UIMax = 1;
    float UIStep = 0.001;
    string UIName = "Offset";
> = 0.04;

/************** light info **************/

half4 light1Pos : POSITION
<
	string UIName = "Light Position";
	string Object = "PointLight";
	string Space = "World";
	int refID = 0;
> = {100.0f, 100.0f, 100.0f, 0.0f};

half4 light1Color : LIGHTCOLOR
<
	int LightRef = 0;
> = { 1.0f, 1.0f, 1.0f, 0.0f };

/************** lighting textures **************/



/***********************************************/
/*** automatically-tracked "tweakables" ********/
/***********************************************/

half4x4 wvp : WorldViewProj < string UIWidget = "None"; >;
half4x4 worldI : WorldI < string UIWidget = "None"; >;
half4x4 viewInvTrans : ViewInverse < string UIWidget = "None"; >;
half4x4 world : World < string UIWidget = "None"; >;

/****************************************************/
/********** CG SHADER FUNCTIONS *********************/
/****************************************************/

// input from application
	struct a2v {
	half4 position		: POSITION;
	half2 texCoord		: TEXCOORD0;
	half3 tangent		: TANGENT;
	half3 binormal		: BINORMAL;
	half3 normal		: NORMAL;
};

// output to fragment program
struct v2f {
        half4 position    	: POSITION;
		half2 texCoord    	: TEXCOORD0;
        float3 eyeVec     	: TEXCOORD1;
        float3 lightVec   	: TEXCOORD2;
		half3 worldNormal	: TEXCOORD3; 
		half3 worldTangent	: TEXCOORD4;
		half3 worldIncident	: TEXCOORD5;
};

// blinn lighting with lit function
half4 blinn2(half3 N,
		half3 L,
		half3 V,
		uniform half4 diffuseColor,
		uniform half4 specularColor,
		uniform half shininess
		)
	{
	half3 H = normalize(V+L);
	half4 lighting = lit(dot(L,N), dot(H,N), shininess);
	return diffuseColor*lighting.y + specularColor*lighting.z;
	}

/**************************************/
/***** VERTEX SHADER ******************/
/**************************************/

v2f v(a2v In,
		uniform half4x4 worldViewProj, // object to clip space
		uniform half4x4 WorldIMatrix,  //world to object space
        uniform half4 lightPosition,
        uniform half4x4 ViewInvTrans,
        uniform half4x4 world        // object to world space
        )
{
	v2f Out;

	Out.worldNormal = mul(In.normal, world).xyz;
	Out.worldTangent = mul(In.tangent, world).xyz;

  // transform vertex position to homogeneous clip space
	Out.position = mul(In.position, worldViewProj);
  
  //pass texture coordinates
	Out.texCoord = In.texCoord;

  // compute the 3x3 tranform from tangent space to object space
	half3x3 objTangentXf;

	objTangentXf[0] = In.binormal.xyz;
	objTangentXf[1] = -In.tangent.xyz;
	objTangentXf[2] = In.normal.xyz;

   //put the world space light position in object space
   	half4 objSpaceLightPos = mul(lightPosition, WorldIMatrix);
   	half3 objLightVec = objSpaceLightPos.xyz - In.position.xyz;
   // xform light vector from obj space to tangent space
   	Out.lightVec = mul(objTangentXf, objLightVec );
	
  //compute the eye vector in world space and put it in object space
  	half4 objSpaceEyePos = mul(ViewInvTrans[3], WorldIMatrix);
  // xform eye vector from obj space to tangent space
  	half3 objEyeVec = objSpaceEyePos.xyz - In.position.xyz;
  	Out.eyeVec = mul(objTangentXf, objEyeVec);
  	
  	//compute the world space incident vector
  	half3 worldSpacePos = mul(In.position, world);
  	Out.worldIncident = worldSpacePos - ViewInvTrans[3];

	return Out;
}

/**************************************/
/***** FRAGMENT PROGRAM ***************/
/**************************************/

float4 f(v2f In,
			uniform sampler2D colorTex,
            uniform sampler2D bumpTex,
            uniform samplerCUBE reflectCube,
            uniform samplerCUBE diffuseCube,
            uniform half4 ambient,
            uniform half4 diffuseColor,
            uniform half4 specularColor,
            uniform half shininess,
            uniform half4 light1Color
            ) : COLOR
{
  // calculate lighting vectors
  half3 V = normalize(In.eyeVec);
  half3 L = normalize(In.lightVec);
  
    //I have to do this take make it look right - don't know why
  half2 newV = V.xy;
  newV.y = -newV.y;
  
  //offset the texture coords based on the view vector
  half height = tex2D(bumpTex, In.texCoord.xy).a;
  half2 offset = newV * (height * 2.0 - 1.0) * offset;
  half2 newTexCoord = In.texCoord.xy + offset;

  //fetch the diffuse map
  half4 colorMap = tex2D(colorTex, newTexCoord);
  
  // fetch the bump normal from the normal map
  half3 normal = tex2D(bumpTex, newTexCoord).xyz * 2.0 - 1.0;
  
    half3 N = normal;

  //calculate attenuation
  float d = length(In.lightVec);

  //create the world normal
    half3 Nn = normalize(In.worldNormal);
    half3 Tn = normalize(In.worldTangent);
    half3 Bn = cross(Nn,Tn);

	half3 worldNorm = Nn + (normal.x * Tn + normal.y * Bn);
  
  //calculate the reflection vector
  half3 refvector = reflect(-In.worldIncident,worldNorm);
  
  //adjust lerp value
  //I'm doing this because I'm lerping vectors instead of vector angles
  half newshininess2 = pow(colorMap.a, 4.5);
  
  //find the new reflection vector based on how shiny the surface is
  //Lerping the angles would be the correct way to do it, but it would be too slow
  //so this is just a fast approximation
  half3 newrefvector = lerp(-worldNorm, refvector, newshininess2);
  
  //I'm doing this because the cube maps were created from a mirror ball.
  //Comment this line out if you're cube map was created another way. 
  newrefvector.y = -newrefvector.y;
  
  //use the new reflection vector to look up the reflection in the diffuse cube and reflection cube
  half4 ref = texCUBE(reflectCube, newrefvector) * specularColor;
  half4 dif = texCUBE(diffuseCube, newrefvector) * diffuseColor;
  
  //combine the two reflection values based on how shiny the surface is
  half4 reflectionColor = lerp(dif,ref,newshininess2);
  
  //lighting
  
  //ambient light
  half4 C = reflectionColor*ambient*(colorMap);
  
  //key light
  C += light1Color * blinn2(N, L, V, (colorMap)*diffuseColor, specularColor*colorMap.a, shininess);

  return C;
}

/****************************************************/
/********** SAMPLERS ********************************/
/****************************************************/

sampler2D colorTextureSampler = sampler_state
{
	Texture = <colorTexture>;
	MinFilter = Linear;
	MagFilter = Linear;
	MipFilter = Anisotropic;
};

samplerCUBE reflectCubeSampler = sampler_state
{
    Texture = <reflectCube>;
    MinFilter = Linear;
    MagFilter = Linear;
    MipFilter = Linear;
};

samplerCUBE diffuseCubeSampler = sampler_state
{
    Texture = <diffuseCube>;
    MinFilter = Linear;
    MagFilter = Linear;
    MipFilter = Linear;
};

sampler2D normalMapSampler = sampler_state
{
	Texture = <normalMap>;
	MinFilter = Linear;
	MagFilter = Linear;
	MipFilter = Linear;
};

/****************************************************/
/********** TECHNIQUES ******************************/
/****************************************************/

technique Complete
{ 
    pass envPass 
    {		
	VertexShader = compile vs_1_1 v(wvp,worldI,light1Pos,viewInvTrans,world);
	ZEnable = true;
	ZWriteEnable = true;
	CullMode = CW;
	AlphaBlendEnable = false;
	PixelShader = compile ps_2_0 f(colorTextureSampler,normalMapSampler,reflectCubeSampler,diffuseCubeSampler,ambient,surfColor,specularColor,shininess,light1Color);
    }
}
